/*	$Id: differentials.c 2007/08/22 18:30:13 mallet $ */

/* 
 * Copyright (c) 2003-2007 LAAS/CNRS                  --  Tue Apr 15 2003
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
static const char *id = "$Id: differentials.c 2007/08/22 18:30:13 mallet $";

/*
 * Created to manage differential transformations
 * These transformation requiere a covariance
 * management system different fron a classical t3d one
 * for when they are applied once again
 *
 *
 * Indeed, if you have defined BaseToOrigin_k^i at time k
 * and you have a position  mainToOrigin_k at time k,
 * then, if i is the best estimator at time k+1 and is a
 * motion estimator (it does only differential motion estimation)
 * then you may think you can deduce V(mainToOrigin_(k+1)) by :
 *
 *    V(mainToOrigin_(k+1))=  J V(mainToOrigin^i_(k+1)) J^T
 *                          + J V(BaseToOrigin_k^i) J^T
 *
 *
 * as you have mainToOrigin_(k+1) = mainToOrigin^i_(k+1) o BaseToOrigin_k^i
 *
 * But, but, but there is a problem. This is incorrect, as there is a
 * correlated part inside both terms of the right side : they are correlated.
 *
 * A fu..ing problem.
 *
 *
 * But Moutarlier has a solution. Yes. A solution.
 * In 2D. But we'll try to test it in 3D
 *
 * It consists in removing the correlation part.
 *
 */

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "t3d/t3d.h"
#include "libpom/lib.h"

#include "pomInt.h"


/**
 * Composes an old BaseToOrigin to estimator me at time index
 */

int
pomIntegrateBaseToOrigin(T3D *m2o, PomMESet *set, int me, int index)
{
  double Joo[36]; /** To simplify o : mainToOrigin^i and g : mainToOrigin at k+1 */
  double Jgt[36];
  double Jgo[36];
  double Jto[36],JRa[9],JRb[9],JRg[9],Jgtr[36];
  double dx,dy,dz,omega,rx[3],ry[3],rz[3],t[16];
  double m1[36],m2[36],m3[36],m4[36];
  int i,j,k,lastbef;

  /** Test the last integration : is it a newer one ? */
  /** Si ce truc est vrai, cela veut dire que l'index pour lequel
   on veut faire une fusion correspond  une date pour laquelle on a
   fait une fusion POSTRIEURE !
  */
  if (pomDateIndex(set,set->me[me].lastabs.date) <= index) {

    /** ON va donc rechercher la meilleure pose antrieure */
    lastbef=pomMELastPos(set,0,index+1);

    /** Si on trouve pas de pose */
    if (lastbef<0) {
      printf ("pomIntegrateBaseToOrigin: no prev pos\n");
      return 1;
    }
    /** Sinon, on intgre  partir de celle-l */
    else {
      t3dCopyIn(m2o,&set->me[0].p[lastbef]->pos.pos.abs.mainToOrigin);
      pomSingleUpdateB2O(m2o,set,me,lastbef);
      if (pomIntegrateBaseToOrigin(m2o,set,me,index)) {
	return 1;
      }
      t3dConvertTo(T3D_U_BRYAN,m2o);
    }
#ifdef POM_DEBUG
    printf ("pomIntegrateBaseToOrigin\n");
    t3dFPrint(stdout, m2o);
#endif
  } else {

    /** Conversion into Euler transforms, first */
    t3dConvertTo(T3D_U_BRYAN,&set->me[me].p[index]->pos.pos.abs.mainToOrigin);
    t3dConvertTo(T3D_U_BRYAN,&set->me[me].baseToOrigin);
    t3dConvertTo(T3D_U_BRYAN,&set->me[me].lastabs.mainToOrigin);
    pomMatrixOfEuler(set->me[me].baseToOrigin.euler.euler,t);

    /** Covariance */
    /** Get  Jgt  */
    pomDerivativeOfEuler(set->me[me].baseToOrigin.euler.euler, JRa,JRb,JRg);

    pom_dm_mult(JRa,set->me[me].p[index]->pos.pos.abs.mainToOrigin.euler.euler+3,rx,3,3,3,1);
    pom_dm_mult(JRb,set->me[me].p[index]->pos.pos.abs.mainToOrigin.euler.euler+3,ry,3,3,3,1);
    pom_dm_mult(JRg,set->me[me].p[index]->pos.pos.abs.mainToOrigin.euler.euler+3,rz,3,3,3,1);
    pom_dm_fill(Jgt,6,6,
	    1.0  ,0.0  ,0.0  ,0.0,0.0,0.0,
	    0.0  ,1.0  ,0.0  ,0.0,0.0,0.0,
	    0.0  ,0.0  ,1.0  ,0.0,0.0,0.0,
	    rx[0],ry[0],rz[0],1.0,0.0,0.0,
	    rx[1],ry[1],rz[1],0.0,1.0,0.0,
	    rx[2],ry[2],rz[2],0.0,0.0,1.0);

    /** Get  Jgo' */
    pom_dm_fill(Jgo,6,6,
	    1.0,0.0,0.0,0.0,0.0,0.0,
	    0.0,1.0,0.0,0.0,0.0,0.0,
	    0.0,0.0,1.0,0.0,0.0,0.0,
	    0.0,0.0,0.0,t[0],t[1],t[2] ,
	    0.0,0.0,0.0,t[4],t[5],t[6] ,
	    0.0,0.0,0.0,t[8],t[9],t[10]);


    /** Get  Jo'o */
    dx=set->me[me].p[index]->pos.pos.abs.mainToOrigin.euler.euler[3]-
      set->me[me].lastabs.mainToOrigin.euler.euler[3];
    dy=set->me[me].p[index]->pos.pos.abs.mainToOrigin.euler.euler[4]-
      set->me[me].lastabs.mainToOrigin.euler.euler[4];
    dz=set->me[me].p[index]->pos.pos.abs.mainToOrigin.euler.euler[5]-
      set->me[me].lastabs.mainToOrigin.euler.euler[5];

    /*
     * On reprend les idees du 2D (approximation par arc de cercle)
     * a relire .... prendre en compte les variations sur les
     * 2 autres angles (1 et 2)
     */
    omega = 0.5*(set->me[me].p[index]->pos.pos.abs.mainToOrigin.euler.euler[0]+
		 set->me[me].lastabs.mainToOrigin.euler.euler[0]);
    pom_dm_fill(Joo,6,6,
		1.0,0.0                         ,0.0,0.0,0.0,0.0,
		0.0,1.0                         ,0.0,0.0,0.0,0.0,
		0.0,0.0                         ,1.0,0.0,0.0,0.0,
		-dy,dz*cos(omega)               ,0.0,1.0,0.0,0.0,
		dx ,dz*sin(omega)               ,0.0,0.0,1.0,0.0,
		0.0,-dx*cos(omega)-dy*sin(omega),0.0,0.0,0.0,1.0);

    /** Get  Jto   */
    pom_dm_mult(JRa,set->me[me].lastabs.mainToOrigin.euler.euler+3,rx,3,3,3,1);
    pom_dm_mult(JRb,set->me[me].lastabs.mainToOrigin.euler.euler+3,ry,3,3,3,1);
    pom_dm_mult(JRg,set->me[me].lastabs.mainToOrigin.euler.euler+3,rz,3,3,3,1);

    /** To be checked i mot sure */
    pom_dm_fill(Jto,6,6,
	    -1.0,0.0,0.0,0.0,0.0,0.0,
	    0.0,-1.0,0.0,0.0,0.0,0.0,
	    0.0,0.0,-1.0,0.0,0.0,0.0,
	    rx[0],ry[0],rz[0],-t[0],-t[1],-t[2] ,
	    rx[1],ry[1],rz[1],-t[4],-t[5],-t[6] ,
	    rx[2],ry[2],rz[2],-t[8],-t[9],-t[10]);

    pom_dmfsf_jmult_6x6(Jgt,set->me[me].baseToOrigin.euler.v, m1);
    pom_dmfsf_jmult_6x6(Jgo,set->me[me].p[index]->pos.pos.abs.mainToOrigin.euler.v, m2);
    pom_dmsf_multtr_6x6(set->me[me].lastabs.mainToOrigin.euler.v,Jto,m3);
    pom_dm_mult(Joo,m3,m4,6,6,6,6);
    pom_dm_mult(Jgo,m4,m3,6,6,6,6);
    pom_dm_trans_6x6(Jgt,Jgtr);
    pom_dm_mult(m3,Jgtr,m4,6,6,6,6);
    pom_dm_plustr_6x6(m4,m3);
    pom_dms_plus_6x6(m1,m2,m2);
    pom_dms_plus_6x6(m2,m3,m4);

    /** Come back to matrix representation */
    t3dConvertTo(T3D_U_MATRIX,&set->me[me].lastabs.mainToOrigin);
    t3dConvertTo(T3D_U_MATRIX,&set->me[me].p[index]->pos.pos.abs.mainToOrigin);
    t3dConvertTo(T3D_U_MATRIX,&set->me[me].baseToOrigin);
    t3dConvertTo(T3D_U_MATRIX,m2o);
    /** Last, the values composition */
    for(i=0;i<4;i++)
      for(j=0;j<4;j++)
	for (k=0, m2o->matrix.matrix[i*4+j]=0.0;k<4;k++)
	  m2o->umatrix.matrix[i*4+j] += set->me[me].baseToOrigin.umatrix.matrix[i*4+k]*set->me[me].p[index]->pos.pos.abs.mainToOrigin.umatrix.matrix[j+k*4];

    for (i=0;i<21;i++)
      m2o->umatrix.v[i]=m4[i];
  }
  return 0;
}
